# -*- coding: utf-8 -*-
# Generated by Django 1.11.23 on 2019-11-15 13:55
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


def add_and_set_default_interlock_category(apps, schema_editor):
    InterlockCardCategory = apps.get_model("NEMO", "InterlockCardCategory")
    InterlockCard = apps.get_model("NEMO", "InterlockCard")
    stanford_category = InterlockCardCategory.objects.create(id=1, name="Stanford", key="stanford")
    InterlockCardCategory.objects.create(id=2, name="WebRelayHttp", key="web_relay_http")
    for interlock_card in InterlockCard.objects.all():
        interlock_card.category = stanford_category
        interlock_card.save()

class Migration(migrations.Migration):

    dependencies = [
        ('NEMO', '0009_version_1_19_0'),
    ]

    operations = [
        migrations.CreateModel(
            name='InterlockCardCategory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=200, help_text="The name for this interlock category")),
                ('key', models.CharField(max_length=100, help_text="The key to identify this interlock category by in interlocks.py")),
            ],
            options={
                'verbose_name_plural': 'Interlock card categories',
                'ordering': ['name'],
            },
        ),
        migrations.AddField(
            model_name='interlockcard',
            name='category',
            field=models.ForeignKey(default=1, on_delete=django.db.models.deletion.CASCADE, to='NEMO.InterlockCardCategory'),
        ),
        migrations.AddField(
            model_name='interlockcard',
            name='password',
            field=models.CharField(blank=True, max_length=100, null=True),
        ),
        migrations.AddField(
            model_name='interlockcard',
            name='username',
            field=models.CharField(blank=True, max_length=100, null=True),
        ),
        migrations.AddField(
            model_name='interlockcard',
            name='enabled',
            field=models.BooleanField(default=True),
        ),
        migrations.AlterField(
            model_name='interlockcard',
            name='even_port',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='interlockcard',
            name='odd_port',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='interlockcard',
            name='number',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='interlock',
            name='channel',
            field=models.PositiveIntegerField(blank=True, null=True, verbose_name='Channel/Relay'),
        ),
        migrations.AlterField(
            model_name='tool',
            name='grant_physical_access_level_upon_qualification',
            field=models.ForeignKey(blank=True,
                                    help_text='The designated physical access level is granted to the user upon qualification for this tool.',
                                    null=True, on_delete=django.db.models.deletion.PROTECT,
                                    to='NEMO.PhysicalAccessLevel'),
        ),
        migrations.AlterField(
            model_name='door',
            name='interlock',
            field=models.OneToOneField(on_delete=django.db.models.deletion.PROTECT, to='NEMO.Interlock'),
        ),
        migrations.AlterField(
            model_name='reservation',
            name='descendant',
            field=models.OneToOneField(blank=True,
                                       help_text='Any time a reservation is moved or resized, the old reservation is cancelled and a new reservation with updated information takes its place. This field links the old reservation to the new one, so the history of reservation moves & changes can be easily tracked.',
                                       null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='ancestor',
                                       to='NEMO.Reservation'),
        ),
        migrations.AlterField(
            model_name='user',
            name='preferences',
            field=models.OneToOneField(null=True, on_delete=django.db.models.deletion.SET_NULL,
                                       to='NEMO.UserPreferences'),
        ),
        migrations.RunPython(add_and_set_default_interlock_category),
    ]



from datetime import timedelta

import requests
from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.db.models import Q
from django.shortcuts import render
from django.urls import Resolver404, resolve
from django.utils import timezone
from django.views.decorators.http import require_GET

from NEMO.models import Alert, LandingPageChoice, Reservation, Resource, UsageEvent
from NEMO.views.alerts import delete_expired_alerts
from NEMO.views.area_access import able_to_self_log_in_to_area, able_to_self_log_out_of_area
from NEMO.views.notifications import delete_expired_notifications, get_notificaiton_counts


@login_required
@require_GET
def landing(request):
	delete_expired_alerts()
	delete_expired_notifications()
	usage_events = UsageEvent.objects.filter(operator=request.user.id, end=None).prefetch_related('tool', 'project')
	tools_in_use = [u.tool.tool_or_parent_id() for u in usage_events]
	fifteen_minutes_from_now = timezone.now() + timedelta(minutes=15)
	landing_page_choices = LandingPageChoice.objects.all()
	if request.device == 'desktop':
		landing_page_choices = landing_page_choices.exclude(hide_from_desktop_computers=True)
	if request.device == 'mobile':
		landing_page_choices = landing_page_choices.exclude(hide_from_mobile_devices=True)
	if not request.user.is_staff and not request.user.is_superuser and not request.user.is_technician:
		landing_page_choices = landing_page_choices.exclude(hide_from_users=True)

	if not settings.ALLOW_CONDITIONAL_URLS:
		# validate all urls
		landing_page_choices = [landing_page_choice for landing_page_choice in landing_page_choices if valid_url_for_landing(landing_page_choice.url)]

	dictionary = {
		'now': timezone.now(),
		'alerts': Alert.objects.filter(Q(user=None) | Q(user=request.user), debut_time__lte=timezone.now()),
		'usage_events': usage_events,
		'upcoming_reservations': Reservation.objects.filter(user=request.user.id, end__gt=timezone.now(), cancelled=False, missed=False, shortened=False).exclude(tool_id__in=tools_in_use, start__lte=fifteen_minutes_from_now).order_by('start')[:3],
		'disabled_resources': Resource.objects.filter(available=False),
		'landing_page_choices': landing_page_choices,
		'notification_counts': get_notificaiton_counts(request.user),
		'self_log_in': able_to_self_log_in_to_area(request.user),
		'self_log_out': able_to_self_log_out_of_area(request.user),
	}
	return render(request, 'landing.html', dictionary)


def valid_url_for_landing(url) -> bool:
	if url.startswith("/"):
		# Internal URL. let's check if it resolves
		try:
			resolve(url)
		except Resolver404:
			return False
	else:
		# External URL, let's check if it exists
		try:
			response = requests.head(url, timeout=0.3)
			if response.status_code > 400:
				return False
		except Exception:
			return False
	return True



import calendar
from datetime import datetime, timedelta

from django.test import TestCase
from django.urls import reverse
from django.utils import timezone

from NEMO.models import ScheduledOutage, Tool, User, Reservation
from NEMO.tests.test_utilities import login_as_staff, login_as_user, test_response_is_landing_page
from NEMO.utilities import localize


class OutageTestCase(TestCase):
	tool = None
	owner = None

	def setUp(self):
		global tool, owner
		owner = User.objects.create(username='mctest', first_name='Testy', last_name='McTester')
		tool = Tool.objects.create(name='test_tool', primary_owner=owner)

	@staticmethod
	def get_outage_data(title='', start: datetime = None, end: datetime = None, tool_name: str = None, outage: bool = False, frequency: str = None, interval: int = None, until: datetime = None):
		if not start:
			start = datetime.now()
		if not end:
			end = start.replace(hour=start.hour + 1)
		data = {
			'title': title,
			'start': calendar.timegm(start.utctimetuple()),
			'end': calendar.timegm(end.utctimetuple()),
			'tool_name': tool_name,
			'recurring_outage': 'on' if outage else '',
			'recurrence_frequency': frequency,
			'recurrence_interval': interval,
			'recurrence_until': localize(until).strftime('%m/%d/%Y') if until else ''
		}
		return data

	def test_outage_policy_problems(self):
		start = datetime.now()
		end = start - timedelta(hours=1)
		data = self.get_outage_data(start=start, end=end, tool_name=tool.name)

		# regular user should not be able to create outage
		login_as_user(self.client)
		response = self.client.get(reverse('create_outage'), {}, follow=True)
		test_response_is_landing_page(self, response)
		# back to staff mode
		login_as_staff(self.client)

		response = self.client.post(reverse('create_outage'), data, follow=True)
		self.assertEquals(response.status_code, 400)
		self.assertEquals(response.content.decode(), "The request parameters have an end time that precedes the start time.")

		# fix time
		end = (start + timedelta(hours=1))
		# Create a reservation and try to schedule an outage at the same time
		Reservation.objects.create(user=owner, creator=owner, tool=tool, start=start, end=end, short_notice=False)
		data = self.get_outage_data(start=start, end=end, tool_name=tool.name)
		response = self.client.post(reverse('create_outage'), data, follow=True)
		self.assertEquals(response.status_code, 400)
		self.assertEquals(response.content.decode(), "Your scheduled outage coincides with a reservation that already exists. Please choose a different time.")

		# try to schedule an outage that starts before but ends slightly after the reservation starts
		data = self.get_outage_data(start=start-timedelta(hours=1), end=end-timedelta(minutes=59), tool_name=tool.name)
		response = self.client.post(reverse('create_outage'), data, follow=True)
		self.assertEquals(response.status_code, 400)
		self.assertEquals(response.content.decode(), "Your scheduled outage coincides with a reservation that already exists. Please choose a different time.")

		# try to schedule an outage that starts slightly before the reservation ends
		data = self.get_outage_data(start=start + timedelta(minutes=59), end=end + timedelta(hours=1), tool_name=tool.name)
		response = self.client.post(reverse('create_outage'), data, follow=True)
		self.assertEquals(response.status_code, 400)
		self.assertEquals(response.content.decode(), "Your scheduled outage coincides with a reservation that already exists. Please choose a different time.")

		# no title
		start = start + timedelta(hours=2)
		end = end + timedelta(hours=2)
		data = self.get_outage_data(start=start, end=end, tool_name=tool.name)
		response = self.client.post(reverse('create_outage'), data, follow=True)
		self.assertEquals(response.status_code, 200) # response code valid but form is sent back. let's make sure the outage was indeed NOT created
		self.assertEquals(ScheduledOutage.objects.all().count(), 0)


	def test_create_outage(self):
		start = datetime.now()
		end = start + timedelta(hours=1)
		data = self.get_outage_data(title="Outage", start=start, end=end, tool_name=tool.name)
		login_as_staff(self.client)
		response = self.client.post(reverse('create_outage'), data, follow=True)
		self.assertEquals(response.status_code, 200)
		self.assertEquals(ScheduledOutage.objects.all().count(), 1)
		self.assertTrue(ScheduledOutage.objects.get(title="Outage"))


	def test_resize_outage(self):
		# create outage
		start = datetime.now()
		end = start + timedelta(hours=1)
		data = self.get_outage_data(title="Outage", start=start, end=end, tool_name=tool.name)
		login_as_staff(self.client)
		response = self.client.post(reverse('create_outage'), data, follow=True)
		self.assertEquals(response.status_code, 200)
		outage = ScheduledOutage.objects.get(title="Outage")
		self.assertTrue(outage.id)

		# regular user should not be able to resize outage
		login_as_user(self.client)
		response = self.client.get(reverse('resize_outage'), {}, follow=True)
		test_response_is_landing_page(self, response)
		self.assertTrue(ScheduledOutage.objects.get(pk=outage.id).id, outage.id)
		# back to staff mode
		login_as_staff(self.client)

		# test wrong delta
		response = self.client.post(reverse('resize_outage'), {'delta':'asd', 'id': outage.id}, follow=True)
		self.assertEquals(response.status_code, 400)
		self.assertEquals(response.content.decode(), "Invalid delta")

		# test no outage id
		response = self.client.post(reverse('resize_outage'), {'delta': 10}, follow=True)
		self.assertEquals(response.status_code, 404)
		self.assertEquals(response.content.decode(), "The outage that you wish to modify doesn't exist!")

		# test resize to less than original time
		response = self.client.post(reverse('resize_outage'), {'delta': -60, 'id': outage.id}, follow=True)
		self.assertEquals(response.status_code, 400)
		self.assertTrue('Outage start time' in response.content.decode())
		self.assertTrue('must be before the end time' in response.content.decode())

		# create a reservation and try to resize outage to overlap reservation
		start_reservation = end + timedelta(hours=1)
		end_reservation = start_reservation + timedelta(hours=1)
		Reservation.objects.create(user=owner, creator=owner, tool=tool, start=start_reservation, end=end_reservation, short_notice=False)
		response = self.client.post(reverse('resize_outage'), {'delta': 61, 'id': outage.id}, follow=True)
		self.assertEquals(response.status_code, 400)
		self.assertEquals(response.content.decode(), "Your scheduled outage coincides with a reservation that already exists. Please choose a different time.")

		# test reduce outage time by 10 min
		response = self.client.post(reverse('resize_outage'), {'delta': -10, 'id':outage.id}, follow=True)
		self.assertEquals(response.status_code, 200)
		self.assertEquals(ScheduledOutage.objects.get(pk=outage.id).end, outage.end - timedelta(minutes=10))

		# test increase outage time by 10 min
		outage = ScheduledOutage.objects.get(pk=outage.id)
		response = self.client.post(reverse('resize_outage'), {'delta': 10, 'id': outage.id}, follow=True)
		self.assertEquals(response.status_code, 200)
		self.assertEquals(ScheduledOutage.objects.get(pk=outage.id).end, outage.end + timedelta(minutes=10))

	def test_move_outage(self):
		# create outage
		start = datetime.now()
		end = start + timedelta(hours=1)
		data = self.get_outage_data(title="Outage", start=start, end=end, tool_name=tool.name)
		login_as_staff(self.client)
		response = self.client.post(reverse('create_outage'), data, follow=True)
		self.assertEquals(response.status_code, 200)
		outage = ScheduledOutage.objects.get(title="Outage")
		self.assertTrue(outage.id)

		# regular user should not be able to move outage
		login_as_user(self.client)
		response = self.client.get(reverse('move_outage'), {}, follow=True)
		test_response_is_landing_page(self, response)
		self.assertTrue(ScheduledOutage.objects.get(pk=outage.id).id, outage.id)
		# back to staff mode
		login_as_staff(self.client)

		# test wrong delta
		response = self.client.post(reverse('move_outage'), {'delta': 'asd', 'id': outage.id}, follow=True)
		self.assertEquals(response.status_code, 400)
		self.assertEquals(response.content.decode(), "Invalid delta")

		# test no outage id
		response = self.client.post(reverse('move_outage'), {'delta': 10}, follow=True)
		self.assertEquals(response.status_code, 404)
		self.assertEquals(response.content.decode(), "The outage that you wish to modify doesn't exist!")

		# create a reservation and try to move outage to overlap reservation
		start_reservation = end + timedelta(hours=1)
		end_reservation = start_reservation + timedelta(hours=1)
		Reservation.objects.create(user=owner, creator=owner, tool=tool, start=start_reservation, end=end_reservation,
								   short_notice=False)
		response = self.client.post(reverse('move_outage'), {'delta': 61, 'id': outage.id}, follow=True)
		self.assertEquals(response.status_code, 400)
		self.assertEquals(response.content.decode(), "Your scheduled outage coincides with a reservation that already exists. Please choose a different time.")

		# test move outage 10 min earlier
		response = self.client.post(reverse('move_outage'), {'delta': -10, 'id': outage.id}, follow=True)
		self.assertEquals(response.status_code, 200)
		self.assertEquals(ScheduledOutage.objects.get(pk=outage.id).end, outage.end - timedelta(minutes=10))
		self.assertEquals(ScheduledOutage.objects.get(pk=outage.id).start, outage.start - timedelta(minutes=10))

		# test move outage 10 min later
		outage = ScheduledOutage.objects.get(pk=outage.id)
		response = self.client.post(reverse('move_outage'), {'delta': 10, 'id': outage.id}, follow=True)
		self.assertEquals(response.status_code, 200)
		self.assertEquals(ScheduledOutage.objects.get(pk=outage.id).end, outage.end + timedelta(minutes=10))
		self.assertEquals(ScheduledOutage.objects.get(pk=outage.id).start, outage.start + timedelta(minutes=10))

	def test_cancel_outage(self):
		# create outage
		start = datetime.now()
		end = start + timedelta(hours=1)
		data = self.get_outage_data(title="Outage", start=start, end=end, tool_name=tool.name)
		login_as_staff(self.client)
		response = self.client.post(reverse('create_outage'), data, follow=True)
		self.assertEquals(response.status_code, 200)
		outage = ScheduledOutage.objects.get(title="Outage")
		self.assertTrue(outage.id)

		# regular user should not be able to delete outage
		login_as_user(self.client)
		response = self.client.get(reverse('cancel_outage', kwargs={'outage_id': 999}), {}, follow=True)
		test_response_is_landing_page(self, response)
		self.assertTrue(ScheduledOutage.objects.get(pk=outage.id).id, outage.id)
		login_as_staff(self.client)

		# get should fail
		response = self.client.get(reverse('cancel_outage', kwargs={'outage_id': 999}), {}, follow=True)
		self.assertEquals(response.status_code, 405)

		# test wrong id
		response = self.client.post(reverse('cancel_outage', kwargs={'outage_id':999}), {}, follow=True)
		self.assertEquals(response.status_code, 404)

		response = self.client.post(reverse('cancel_outage', kwargs={'outage_id': outage.id}), {}, follow=True)
		self.assertEquals(response.status_code, 200)
		self.assertEquals(ScheduledOutage.objects.all().count(), 0)

	def test_outage_details(self):
		# create outage
		start = datetime.now()
		end = start + timedelta(hours=1)
		data = self.get_outage_data(title="Outage", start=start, end=end, tool_name=tool.name)
		login_as_staff(self.client)
		response = self.client.post(reverse('create_outage'), data, follow=True)
		self.assertEquals(response.status_code, 200)
		outage = ScheduledOutage.objects.get(title="Outage")
		self.assertTrue(outage.id)

		# anybody that is logged in can see outage details
		login_as_user(self.client)

		# post should fail
		response = self.client.post(reverse('outage_details', kwargs={'outage_id': 999}), {}, follow=True)
		self.assertEquals(response.status_code, 405)

		# test wrong id
		response = self.client.get(reverse('outage_details', kwargs={'outage_id':999}), {}, follow=True)
		self.assertEquals(response.status_code, 404)

		response = self.client.get(reverse('outage_details', kwargs={'outage_id': outage.id}), {}, follow=True)
		self.assertEquals(response.status_code, 200)

	def test_no_tool_name_404(self):
		start = datetime.now()
		end = start + timedelta(hours=1)
		until = datetime.now() + timedelta(days=5)

		data = self.get_outage_data(start=start, end=end, outage=True, frequency='DAILY', interval=1, until=until)

		login_as_staff(self.client)
		response = self.client.post(reverse('create_outage'), data, follow=True)
		self.assertEquals(response.status_code, 404)

	def test_every_day_for_a_week(self):
		start = datetime.now()
		end = start + timedelta(hours=1)
		until = datetime.now() + timedelta(days=6)

		data = self.get_outage_data(title='every day outage week', start=start, end=end, tool_name=tool.name, outage=True, frequency='DAILY', interval=1, until=until)

		login_as_staff(self.client)
		response = self.client.post(reverse('create_outage'), data, follow=True)

		self.assertEquals(response.status_code, 200)
		outages = ScheduledOutage.objects.filter(title='every day outage week', tool=tool)
		self.assertEqual(len(outages), 7)

	def test_every_week_for_a_year(self):
		start = datetime.now().replace(microsecond=0)
		end = start + timedelta(hours=1)
		until = datetime.now() + timedelta(days=365)

		data = self.get_outage_data(title='every day outage year', start=start, end=end, tool_name=tool.name, outage=True, frequency='WEEKLY', interval=1, until=until)

		login_as_staff(self.client)
		response = self.client.post(reverse('create_outage'), data, follow=True)

		self.assertEquals(response.status_code, 200)
		outages = ScheduledOutage.objects.filter(title='every day outage year', tool=tool)
		for outage in outages:
			good_start = outage.start.astimezone(timezone.get_current_timezone())
			good_end = outage.end.astimezone(timezone.get_current_timezone())
			self.assertEquals(good_start.weekday(), start.weekday())
			self.assertEquals(good_end.weekday(), end.weekday())
			self.assertEquals(good_start.time(), start.time())
			self.assertEquals(good_end.time(), end.time())

	def test_week_day(self):
		start = datetime.now()
		end = start + timedelta(hours=1)
		until = datetime.now() + timedelta(weeks=9)

		data = self.get_outage_data(title='every week day outage', start=start, end=end, tool_name=tool.name, outage=True, frequency='DAILY_WEEKDAYS', interval=1, until=until)

		login_as_staff(self.client)
		response = self.client.post(reverse('create_outage'), data, follow=True)

		self.assertEquals(response.status_code, 200)
		outages = ScheduledOutage.objects.filter(title='every week day outage', tool=tool)
		for outage in outages:
			# 0 is Monday, 5 & 6 are Saturday and Sunday
			self.assertLess(outage.start.astimezone(timezone.get_current_timezone()).weekday(), 5)

	def test_weekend(self):
		start = datetime.now()
		end = start + timedelta(hours=1)
		until = datetime.now() + timedelta(weeks=9)

		data = self.get_outage_data(title='every weekend day outage', start=start, end=end, tool_name=tool.name, outage=True, frequency='DAILY_WEEKENDS', interval=1, until=until)

		login_as_staff(self.client)
		response = self.client.post(reverse('create_outage'), data, follow=True)

		self.assertEquals(response.status_code, 200)
		outages = ScheduledOutage.objects.filter(title='every weekend day outage', tool=tool)
		for outage in outages:
			# 0 is Monday, 5 & 6 are Saturday and Sunday
			self.assertGreaterEqual(outage.start.astimezone(timezone.get_current_timezone()).weekday(), 5)



from typing import List

from django.forms import Widget
from django.utils.safestring import mark_safe

from NEMO.models import User, Tool


class ToolTree(Widget):
	def render(self, name, value, attrs=None, renderer=None):
		"""
		This widget takes a list of tools and creates nested unordered lists in a hierarchical manner.
		The parameters name and attrs are not used.
		'value' is a dictionary which must contain a 'tools' key with a value that is a QuerySet of all tools to be put in the list.
		A collection of unordered HTML lists is returned with various callbacks and properties attached to each nested element.

		For a more concrete example, suppose the following tools are input to the tool tree:
		Packaging/Dicing Saw
		Chemical Vapor Deposition/PECVD
		Gen Furnaces/Sinter

		The following unordered HTML list would be produced:
		<ul>
			<li>
				<a href="javascript:void(0);" onclick="on_tool_tree_click($(this.nextSibling))" class="node">Packaging</a>
				<ul class="collapsed">
					<li><a href="javascript:void(0);" onclick="on_tool_tree_click($(this))" class="leaf node">Dicing saw</a></li>
				</ul>
			</li>
			<li>
				<a href="javascript:void(0);" onclick="on_tool_tree_click($(this.nextSibling))" class="node">Chemical Vapor Deposition</a>
				<ul class="collapsed">
					<li><a href="javascript:void(0);" onclick="on_tool_tree_click($(this))" class="leaf node">PECVD</a></li>
				</ul>
			</li>
			<li>
				<a href="javascript:void(0);" onclick="on_tool_tree_click($(this.nextSibling))" class="node">Gen Furnaces</a>
				<ul class="collapsed">
					<li><a href="javascript:void(0);" onclick="on_tool_tree_click($(this))" class="leaf node">Sinter</a></li>
				</ul>
			</li>
		</ul>
		"""
		tree = ToolTreeHelper(None)
		user: User = value['user'] if 'user' in value else None
		tools: List[Tool] = value['tools']
		for tool in tools:
			is_qualified = (user and user.is_staff) or (user and tool in user.qualifications.all())
			tree.add(tool.category + '/' + tool.name_or_child_in_use_name(), tool.id,  is_qualified)
		return mark_safe(tree.render())


class ToolTreeHelper:
	"""
	This class reads in a textual representation of the organization of each NanoFab tool and renders it to equivalent
	unordered HTML lists.
	"""
	def __init__(self, name):
		self.name = name
		self.children = []
		self.id = None
		self.is_user_qualified = False

	def add(self, tool, identifier, is_user_qualified):
		"""
		This function takes as input a string representation of the tool in the organization hierarchy.
		Example input might be "Imaging and Analysis/Microscopes/Zeiss FIB". The input is parsed with '/' as the
		separator and the tool is added to the class' tree structure.
		"""
		part = tool.partition('/')
		for child in self.children:
			if child.name == part[0]:
				child.add(part[2], identifier, is_user_qualified)
				return
		self.children.append(ToolTreeHelper(part[0]))
		if part[2] != '':
			self.children[-1].add(part[2], identifier, is_user_qualified)
		else:
			self.children[-1].id = identifier
			self.children[-1].is_user_qualified = is_user_qualified

	def render(self):
		"""
		This function cycles through the root node of the tool list and enumerates all the child nodes directly.
		The function assumes that a tree structure of the tools has already been created by calling 'add(...)' multiple
		times. A string of unordered HTML lists is returned.
		"""
		result = '<ul class="nav nav-list" id="tool_tree" style="display:none">'
		for child in self.children:
			result += self.__render_helper(child, '')
		result += '</ul>'
		return result

	def __render_helper(self, node, result):
		"""
		Recursively dive through the tree structure and convert it to unordered HTML lists.
		Each node is output as an HTML list item. If the node has children then those are also output.
		"""
		result += '<li>'
		if node.__is_leaf():
			css_class = "" if node.is_user_qualified else 'class="disabled"'
			result += f'<a href="javascript:void(0);" onclick="set_selected_item(this)" data-tool-id="{node.id}" data-type="tool link" {css_class}>{node.name}</a>'
		if not node.__is_leaf():
			result += f'<label class="tree-toggler nav-header"><div>{node.name}</div></label><ul class="nav nav-list tree" data-category="{node.name}">'
			for child in node.children:
				result = self.__render_helper(child, result)
			result += '</ul>'
		result += '</li>'
		return result

	def __is_leaf(self):
		""" Test if this node is a leaf (i.e. an actual tool). If it is not then the node must be a tool category. """
		return self.children == []

	def __str__(self):
		""" For debugging, output a string representation of the object in the form: <name [child1, child2, child3, ...]> """
		result = str(self.name)
		if not self.__is_leaf():
			result += ' [' + ', '.join(str(child) for child in self.children) + ']'
		return result



from socket import gethostbyname_ex, gaierror
from socket import error as socket_error
from _ssl import CERT_REQUIRED, PROTOCOL_TLSv1_2
from getpass import getpass
from os import remove
from ssl import get_server_certificate
from subprocess import run
from sys import argv
from textwrap import dedent

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric.rsa import generate_private_key
from cryptography.hazmat.primitives.hashes import SHA256
from cryptography.hazmat.primitives.serialization import Encoding, NoEncryption, PrivateFormat
from cryptography.x509 import NameAttribute, NameOID, DNSName, CertificateSigningRequestBuilder, Name, SubjectAlternativeName
from django.utils.crypto import get_random_string
from ldap3 import Tls, Server, Connection, AUTO_BIND_TLS_BEFORE_BIND, SIMPLE
from ldap3.core.exceptions import LDAPBindError, LDAPExceptionError


def entry_point():
	request = argv[1] if len(argv) == 2 else None
	if request == 'generate_secret_key':
		generate_secret_key()
	elif request == 'query_public_key':
		query_public_key()
	elif request == 'test_ldap_authentication':
		test_ldap_authentication()
	elif request == 'generate_tls_keys':
		generate_tls_keys()
	elif request == 'install_systemd_service':
		install_systemd_service()
	elif request == 'uninstall_systemd_service':
		uninstall_systemd_service()
	elif request == 'open_firewall_ports':
		open_firewall_ports()
	else:
		usage = \
				"""
				You can provision everything NEMO needs to run with these commands...
				
				generate_secret_key -
					Prints a randomly generated secret key, used for hashing user sessions,
					cryptographic signing, and more. More information is available from the Django documentation:
					https://docs.djangoproject.com/en/1.11/ref/settings/#std:setting-SECRET_KEY
				
				query_public_key -
					Opens a TLS connection to another server and prints the public TLS key.
					This is useful if you want to verify correct TLS configuration,
					or pin a particular certificate for LDAPS authentication.
				
				test_ldap_authentication -
					Opens an encrypted connection (using only TLS 1.2) to an LDAP directory (such as Active Directory)
					and tests authentication. You will be prompted for the hostname of the authentication server,
					the domain name, a username, password, and public TLS key.
				
				generate_tls_keys -
					Generates an RSA 4096 bit private key and certificate signing request. These can be used
					to obtain a public key, enabling TLS encryption for your NEMO users. This keeps communication
					between a user's web browser and the NEMO server secure.
				
				install_systemd_service -
					Installs NEMO as a service with Systemd (by creating a file /etc/systemd/system/nemo.service).
					After running this command you can manipulate NEMO with standard systemctl commands, such as:
					systemctl start nemo
					systemctl stop nemo
					systemctl restart nemo
					systemctl enable nemo
					systemctl disable nemo
					See this tutorial on Systemd for more information:
					https://www.digitalocean.com/community/tutorials/how-to-use-systemctl-to-manage-systemd-services-and-units
				
				uninstall_systemd_service -
					Uninstalls NEMO as a service with Systemd (by simply deleting /etc/systemd/system/nemo.service).
				
				open_firewall_ports -
					Configure the firewall to allow incoming web browser requests.
					Requests received on port 80 are intended to be redirected to port 443 for improved security.
					Port 80 (HTTP) and 443 (HTTPS) are made available via firewall-cmd:
					firewall-cmd --zone=public --permanent --add-service=http
					firewall-cmd --zone=public --permanent --add-service=https
				"""
		print(dedent(usage))


def generate_secret_key():
	"""
	Writes a new Django secret key to the specified file. Use this with your settings.py file for hashing and signing.
	Documentation available at https://docs.djangoproject.com/en/1.11/ref/settings/#std:setting-SECRET_KEY
	"""
	print(get_random_string(50, "abcdefghijklmnopqrstuvwxyz0123456789!@$%&=-+#(*_^)"))


def query_public_key():
	"""
	Opens a secure connection to a server and downloads its public TLS key.
	The key is printed in PEM format.
	"""
	print("Query a server's public TLS key.")
	print("If the DNS name resolves to multiple IP addresses then the public keys for all IP addresses are printed.")
	print("Port 636 is commonly used for secure LDAP.")
	print("Port 443 is commonly used for secure HTTP.")
	name = input("DNS name or IP address = ")
	port = input("Port = ")

	try:
		ip_addresses = gethostbyname_ex(name)[2]
	except gaierror:
		print('DNS name resolution failed.')
		return
	except socket_error as error:
		print(f'Socket error: {error}')
		return

	if len(ip_addresses) == 0:
		print('No IP address was returned from DNS name resolution.')
	elif len(ip_addresses) == 1:
		try:
			certificate = get_server_certificate((name, port))
			print(certificate)
		except socket_error as error:
			print(f'Socket error: {error}')
			return
	elif len(ip_addresses) > 1:
		print(f"Name resolves to {len(ip_addresses)} IP addresses: {ip_addresses}")
		for ip in ip_addresses:
			print(f"Querying certificate for {ip}:")
			try:
				certificate = get_server_certificate((ip, port))
				print(certificate)
			except socket_error as error:
				print(f'Socket error: {error}')


def test_ldap_authentication():
	""" Opens an encrypted connection (using only TLS 1.2) to an LDAP directory (such as Active Directory) and tests authentication """
	name = input("DNS name or IP address = ")
	domain = input("Domain = ")
	username = input("Username = ")
	password = getpass("Password = ")
	certificate = input("Path to public key certificate = ")
	try:
		t = Tls(validate=CERT_REQUIRED, version=PROTOCOL_TLSv1_2, ca_certs_file=certificate)
		s = Server(name, port=636, use_ssl=True, tls=t)
		c = Connection(s, user='{}\\{}'.format(domain, username), password=password, auto_bind=AUTO_BIND_TLS_BEFORE_BIND, authentication=SIMPLE)
		c.unbind()
		# At this point the user successfully authenticated to at least one LDAP server.
		print("Authentication successful!")
	except LDAPBindError:
		pass  # When this error is caught it means the username and password were invalid against the LDAP server.
	except LDAPExceptionError as error:
		print("A problem was encountered during authentication:\n")
		print(error)


def generate_tls_keys():
	""" Creates a TLS private key (RSA, 4096 bits, PEM format) and certificate signing request (CSR). """

	# Query the user for CSR attributes
	country = input("Country: ")
	state = input("State or province: ")
	locality = input("Locality: ")
	organization = input("Organization: ")
	organizational_unit = input("Organizational unit: ")
	email = input("Email: ")
	common_name = input("Common name: ")

	print("Enter any subject alternative names (SANs) you wish to attach to the certificate. Leave blank to continue.")
	sans = []
	while True:
		san = input("Subject alternative name: ")
		if san == '':
			break
		sans.append(DNSName(san))

	# Make sure we can open the output files first
	private_key_file = open("private.key", "wb")
	csr_file = open("certificate_signing_request", "wb")

	# Generate the private key
	key = generate_private_key(public_exponent=65537, key_size=4096, backend=default_backend())

	attributes = [
		NameAttribute(NameOID.COUNTRY_NAME, country),
		NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, state),
		NameAttribute(NameOID.LOCALITY_NAME, locality),
		NameAttribute(NameOID.ORGANIZATION_NAME, organization),
		NameAttribute(NameOID.ORGANIZATIONAL_UNIT_NAME, organizational_unit),
		NameAttribute(NameOID.EMAIL_ADDRESS, email),
		NameAttribute(NameOID.COMMON_NAME, common_name),
	]

	# Generate the CSR and sign it with the private key
	csr = CertificateSigningRequestBuilder().subject_name(Name(attributes))
	if sans:
		csr = csr.add_extension(SubjectAlternativeName(sans), critical=False)
	csr = csr.sign(key, SHA256(), default_backend())

	# Write the private key and CSR to disk
	private_key_file.write(key.private_bytes(encoding=Encoding.PEM, format=PrivateFormat.TraditionalOpenSSL, encryption_algorithm=NoEncryption()))
	csr_file.write(csr.public_bytes(Encoding.PEM))

	private_key_file.close()
	csr_file.close()

	# Success!
	print("Successfully generated a private key and certificate signing request.")


def install_systemd_service():
	""" Installs NEMO as a systemd service. """
	customizations = {
		'user': 'nemo',
		'group': 'nemo',
		'process_identifier_file': '/tmp/nemo.pid',
		'gunicorn': '/home/nemo/python/bin/gunicorn',
	}

	service = """
	[Unit]
	Description=NEMO is a laboratory logistics web application. Use it to schedule reservations, control tool access, track maintenance issues, and more.
	After=network.target syslog.target nss-lookup.target

	[Service]
	User={user}
	Group={group}
	PIDFile={process_identifier_file}
	ExecStart={gunicorn} NEMO.wsgi:application
	ExecReload=/bin/kill -s HUP $MAINPID
	ExecStop=/bin/kill -s TERM $MAINPID
	PrivateTmp=true

	[Install]
	WantedBy=multi-user.target
	""".format(**customizations)

	with open('/etc/systemd/system/nemo.service', 'w') as f:
		f.write(dedent(service))

	print("The systemd NEMO service was successfully installed.")


def uninstall_systemd_service():
	""" Uninstalls NEMO as a systemd service """
	remove('/etc/systemd/system/nemo.service')
	print("The systemd NEMO service was successfully uninstalled.")


def open_firewall_ports():
	http_result = run(['firewall-cmd', '--zone=public', '--permanent', '--add-service=http'])
	https_result = run(['firewall-cmd', '--zone=public', '--permanent', '--add-service=https'])
	reload_result = run(['firewall-cmd', '--reload'])


if __name__ == "__main__":
	entry_point()
